/*-
 * Copyright (c) 1980, 1987, 1988, 1991, 1993, 1994
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * login [ name ]
 * login -h hostname	(for telnetd, etc.)
 * login -f name	(for pre-authenticated login: datakit, xterm, etc.)
 */

#include "bsd_locl.h"

RCSID("$Id: login.c,v 1.76 1996/08/24 20:30:56 assar Exp $");

#ifdef	xSKEY			/* Don't bother to find this header file. */
#include <skey.h>
#endif

#include "sysv_default.h"
#ifdef SYSV_SHADOW
#include "sysv_shadow.h"
#endif

void	 badlogin __P((char *));
void	 checknologin __P((void));
void	 dolastlog __P((int));
void	 getloginname __P((int));
void	 motd __P((void));
int	 rootterm __P((char *));
void	 sigint __P((int));
char	*stypeof __P((char *));
void	 timedout __P((int));
static int	 doremotelogin __P((char *));
void     login_fbtab __P((char *, uid_t, gid_t));
#ifdef KERBEROS
int	 klogin __P((struct passwd *, char *, char *, char *));
#endif

extern void login __P((struct utmp *));

#define	TTYGRPNAME	"tty"		/* name of group to own ttys */

/*
 * This bounds the time given to login.  Not a define so it can
 * be patched on machines where it's too small.
 */
u_int	login_timeout = 300;

#ifdef KERBEROS
int	notickets = 1;
int	noticketsdontcomplain = 1;
char	*instance;
char	*krbtkfile_env;
int	authok;
#endif

char    *ttyprompt;

char *__progname;
struct	passwd *pwd;
int	failures;
char	term[64], *envinit[1], *hostname, *username, *tty;

static  char rusername[100], lusername[100];
#ifdef HAVE_SHADOW_H
static  struct spwd *spwd;
#endif

static int
change_passwd(struct passwd  *who)
{
        int             status;
        int             pid;
        int             wpid;
 
        switch (pid = fork()) {
	case -1:
	    perror("Cannot execute /bin/passwd");
	    sleepexit(1);
	case 0:
	    execlp("/bin/passwd", "passwd", who->pw_name, (char *) 0);
	    _exit(1);
	default:
	    while ((wpid = wait(&status)) != -1 && wpid != pid)
		/* void */ ;
	    return (status);
	}
}

int
main(argc, argv)
	int argc;
	char *argv[];
{
	struct group *gr;
	int ask, ch, cnt, fflag, hflag, pflag, quietlog, rootlogin, rval;
	int rflag;
	int changepass = 0;
	uid_t uid;
	char *domain, *p, passwd[128], ep[20], *salt, *ttyn;
	char tbuf[MaxPathLen + 2], tname[sizeof(_PATH_TTY) + 10];
	char localhost[MaxHostNameLen];
	char full_hostname[MaxHostNameLen];
#ifdef	SKEY
	int permit_passwd = 0;
#endif
	int mask = 022;		/* Default umask (set below) */
	int maxtrys = 5;	/* Default number of allowed failed logins */

	__progname = argv[0];

	openlog("login", LOG_ODELAY, LOG_AUTH);

        /* Read defaults file and set the login timeout period. */
        sysv_defaults();
        login_timeout = atoi(default_timeout);
        maxtrys = atoi(default_maxtrys);
        if (sscanf(default_umask, "%o", &mask) != 1 || (mask & ~0777))
                syslog(LOG_WARNING, "bad umask default: %s", default_umask);
        else
                umask(mask);

	(void)signal(SIGALRM, timedout);
	(void)alarm(login_timeout);
	(void)signal(SIGQUIT, SIG_IGN);
	(void)signal(SIGINT, SIG_IGN);
	(void)setpriority(PRIO_PROCESS, 0, 0);

	/*
	 * -p is used by getty to tell login not to destroy the environment
	 * -f is used to skip a second login authentication
	 * -h is used by other servers to pass the name of the remote
	 *    host to login so that it may be placed in utmp and wtmp
	 * -r is used by old-style rlogind to execute the autologin protocol
	 */
	*full_hostname = '\0';
	domain = NULL;
	if (k_gethostname(localhost, sizeof(localhost)) < 0)
		syslog(LOG_ERR, "couldn't get local hostname: %m");
	else
		domain = strchr(localhost, '.');

	fflag = hflag = pflag = rflag = 0;
	uid = getuid();
	while ((ch = getopt(argc, argv, "d:fh:pr:")) != EOF)
	        switch (ch) {
		case 'd':
                                break;
		case 'f':
			fflag = 1;
			break;
		case 'h':
			if (rflag || hflag) {
			    printf("Only one of -r and -h allowed\n");
			    exit(1);
                        }
			if (uid)
				errx(1, "-h option: %s", strerror(EPERM));
			hflag = 1;
			strncpy(full_hostname, optarg, sizeof(full_hostname)-1);
			if (domain && (p = strchr(optarg, '.')) &&
			    strcasecmp(p, domain) == 0)
				*p = 0;
			hostname = optarg;
			break;
		case 'p':
			if (getuid()) {
			    fprintf(stderr,
                                    "login: -p for super-user only.\n");
			    exit(1);
                        }
			pflag = 1;
			break;
	        case 'r':
			if (rflag || hflag) {
			    printf("Only one of -r and -h allowed\n");
			    exit(1);
                        }
			if (getuid()) {
			    fprintf(stderr,
                                    "login: -r for super-user only.\n");
			    exit(1);
                        }
			rflag = 1;
			strncpy(full_hostname, optarg, sizeof(full_hostname)-1);
			if (domain && (p = strchr(optarg, '.')) &&
			    strcasecmp(p, domain) == 0)
				*p = 0;
			hostname = optarg;
			fflag = (doremotelogin(full_hostname) == 0);
			break;
		case '?':
		default:
			if (!uid)
				syslog(LOG_ERR, "invalid flag %c", ch);
			(void)fprintf(stderr,
			    "usage: login [-fp] [-h hostname | -r hostname] [username]\n");
			exit(1);
		}
	argc -= optind;
	argv += optind;

	if (geteuid() != 0) {
	        fprintf(stderr, "login: only root may use login, use su\n");
	        /* Or install login setuid root, which is not necessary */
		exit(1);
	}
        /*
         * Figure out if we should ask for the username or not. The name
         * may be given on the command line or via the environment, and
         * it may even be in the terminal input queue.
         */
        if (rflag) {
                username = lusername;
                ask = 0;
	      } else
        if (*argv && strchr(*argv, '=')) {
                ask = 1;
	      } else
        if (*argv && strcmp(*argv, "-") == 0) {
                argc--;
                argv++;
                ask = 1;
	      } else
	if (*argv) {
		username = *argv;
		ask = 0;
                argc--;
                argv++;
        } else if ((ttyprompt = getenv("TTYPROMPT")) && *ttyprompt) {
                getloginname(0);
                ask = 0;
	} else
		ask = 1;

	/* Default tty settings. */
	stty_default();

	for (cnt = getdtablesize(); cnt > 2; cnt--)
		(void)close(cnt);

        /*
         * Determine the tty name. BSD takes the basename, SYSV4 takes
         * whatever remains after stripping the "/dev/" prefix. The code
         * below should produce sensible results in either environment.
         */
        ttyn = ttyname(STDIN_FILENO);
        if (ttyn == NULL || *ttyn == '\0') {
	        (void)sprintf(tname, "%s??", _PATH_TTY);
	        ttyn = tname;
	}
        if ((tty = strchr(ttyn + 1, '/')))
	        ++tty;
        else
	        tty = ttyn;

	for (cnt = 0;; ask = 1) {
	        char prompt[128];
		if (ask) {
			fflag = 0;
			getloginname(1);
		}
		rootlogin = 0;
#ifdef	KERBEROS
		if ((instance = strchr(username, '.')) != NULL) {
			if (strncmp(instance, ".root", 5) == 0)
				rootlogin = 1;
			*instance++ = '\0';
		} else
			instance = "";
#endif
		if (strlen(username) > UT_NAMESIZE)
			username[UT_NAMESIZE] = '\0';

		/*
		 * Note if trying multiple user names; log failures for
		 * previous user name, but don't bother logging one failure
		 * for nonexistent name (mistyped username).
		 */
		if (failures && strcmp(tbuf, username)) {
			if (failures > (pwd ? 0 : 1))
				badlogin(tbuf);
			failures = 0;
		}
		(void)strcpy(tbuf, username);

		if ((pwd = k_getpwnam(username)) != 0)
			salt = pwd->pw_passwd;
		else
			salt = "xx";

		if (pwd && pwd->pw_uid == 0 && strcmp("root", username) != 0)
		  {
		    pwd = 0;
		    syslog(LOG_ALERT,
			   "NIS attack, user %s has uid 0", username);
		  }

		/*
		 * if we have a valid account name, and it doesn't have a
		 * password, or the -f option was specified and the caller
		 * is root or the caller isn't changing their uid, don't
		 * authenticate.
		 */
		if (pwd) {
			if (pwd->pw_uid == 0)
				rootlogin = 1;

			if (fflag && (uid == 0 || uid == pwd->pw_uid)) {
				/* already authenticated */
				break;
			} else if (pwd->pw_passwd[0] == '\0') {
				/* pretend password okay */
				rval = 0;
				goto ttycheck;
			}
		}

		fflag = 0;

		(void)setpriority(PRIO_PROCESS, 0, -4);

		sprintf(prompt, "%s's Password: ", username);
#ifdef	SKEY
		permit_passwd = skeyaccess(pwd, tty,
					   hostname ? full_hostname : NULL,
					   NULL);
		p = (char *) skey_getpass(prompt, pwd, permit_passwd);
		strncpy(ep,
			(char *)skey_crypt(p, salt, pwd, permit_passwd),
			sizeof(ep));
		ep[sizeof(ep) - 1] = 0;
#else
		des_read_pw_string (passwd, sizeof(passwd) - 1, prompt, 0);
		passwd[sizeof(passwd) - 1] = '\0';
		strncpy(ep, crypt(passwd, salt), sizeof(ep));
		ep[sizeof(ep) - 1] = 0;
#endif

		if (pwd) {
#ifdef KERBEROS
#ifdef SKEY
			/*
			 * Do not allow user to type in kerberos password
			 * over the net (actually, this is ok for encrypted
			 * links, but we have no way of determining if the
			 * link is encrypted.
			 */
			if (!permit_passwd) {
				rval = 1;		/* failed */
			} else
#endif
			rval = klogin(pwd, instance, localhost, passwd);
			memset(passwd, 0, sizeof(passwd));
			/* getpwnam() is not reentrant and klogin used it! */
			pwd = k_getpwnam(username);
			if (pwd == NULL){
			      fprintf(stderr, "unknown login %s\n", username);
			      exit(1);
			}
			if (rval != 0 && rootlogin && pwd->pw_uid != 0)
				rootlogin = 0;
			if (rval == 0)
				authok = 1;
			else if (rval == 1)
				rval = strcmp(ep, pwd->pw_passwd);
#else
			rval = strcmp(ep, pwd->pw_passwd);
#endif
		}
		memset(passwd, 0, sizeof(passwd));

		(void)setpriority(PRIO_PROCESS, 0, 0);

	ttycheck:
		/*
		 * If trying to log in as root without Kerberos,
		 * but with insecure terminal, refuse the login attempt.
		 */
#ifdef KERBEROS
		if (authok == 0)
#endif
		if (pwd && !rval && rootlogin && !rootterm(tty)
		    && !rootterm(ttyn)) {
			(void)fprintf(stderr,
			    "%s login refused on this terminal.\n",
			    pwd->pw_name);
			if (hostname)
				syslog(LOG_NOTICE,
				    "LOGIN %s REFUSED FROM %s ON TTY %s",
				    pwd->pw_name, hostname, tty);
			else
				syslog(LOG_NOTICE,
				    "LOGIN %s REFUSED ON TTY %s",
				     pwd->pw_name, tty);
			continue;
		}

		if (pwd && !rval)
			break;

		(void)printf("Login incorrect\n");
		failures++;

                /* max number of attemps and delays taken from defaults file */
		/* we allow maxtrys tries, but after 2 we start backing off */
		if (++cnt > 2) {
			if (cnt >= maxtrys) {
				badlogin(username);
				sleepexit(1);
			}
			sleep((u_int)((cnt - 2) * atoi(default_sleep)));
		}
	}

	/* committed to login -- turn off timeout */
	(void)alarm((u_int)0);

	endpwent();

#if defined(HAVE_GETUDBNAM) && defined(HAVE_SETLIM)
	{
	    struct udb *udb;
	    struct resclim lim;
	    int t;
	    const int maxcpu = 46116860184; /* some random constant */
	    udb = getudbnam(pwd->pw_name);
	    if(udb == UDB_NULL){
		fprintf(stderr, "Failed to get UDB entry.\n");
		exit(1);
	    }
	    lim.resc_resource = L_CPU;
	    lim.resc_category = C_PROC;
	    lim.resc_type = L_T_ABSOLUTE;
	    lim.resc_action = 0;
	    lim.resc_used = 0;
	    t = udb->ue_pcpulim[UDBRC_INTER];
	    if(t == 0 || t > maxcpu)
		lim.resc_value[lim.resc_type] = CPUUNLIM;
	    else
		lim.resc_value[lim.resc_type] = 100 * CLOCKS_PER_SEC * t;

	    if(setlim(0, &lim) < 0){
		fprintf(stderr, "setlim: %s\n", strerror(errno));
	    }

	    lim.resc_resource = L_CPU;
	    lim.resc_category = C_SESS;
	    lim.resc_type = L_T_ABSOLUTE;
	    lim.resc_action = 0;
	    lim.resc_used = 0;
	    t = udb->ue_jcpulim[UDBRC_INTER];
	    if(t == 0 || t > maxcpu)
		lim.resc_value[lim.resc_type] = CPUUNLIM;
	    else
		lim.resc_value[lim.resc_type] = 100 * CLOCKS_PER_SEC * t;

	    if(setlim(0, &lim) < 0){
		fprintf(stderr, "setlim: %s\n", strerror(errno));
	    }
	}
#endif
	/* if user not super-user, check for disabled logins */
	if (!rootlogin)
		checknologin();

	if (chdir(pwd->pw_dir) < 0) {
		(void)printf("No home directory %s!\n", pwd->pw_dir);
		if (chdir("/"))
			exit(0);
		pwd->pw_dir = "/";
		(void)printf("Logging in with home = \"/\".\n");
	}

	quietlog = access(_PATH_HUSHLOGIN, F_OK) == 0;

#if defined(HAVE_PASSWD_CHANGE) && defined(HAVE_PASSWD_EXPIRE)
	if (pwd->pw_change || pwd->pw_expire)
		(void)gettimeofday(&tp, (struct timezone *)NULL);

	if (pwd->pw_change)
		if (tp.tv_sec >= pwd->pw_change) {
			(void)printf("Sorry -- your password has expired.\n");
			changepass=1;
		} else if (pwd->pw_change - tp.tv_sec <
		    2 * DAYSPERWEEK * SECSPERDAY && !quietlog)
			(void)printf("Warning: your password expires on %s",
			    ctime(&pwd->pw_change));
	if (pwd->pw_expire)
		if (tp.tv_sec >= pwd->pw_expire) {
			(void)printf("Sorry -- your account has expired.\n");
			sleepexit(1);
		} else if (pwd->pw_expire - tp.tv_sec <
		    2 * DAYSPERWEEK * SECSPERDAY && !quietlog)
			(void)printf("Warning: your account expires on %s",
			    ctime(&pwd->pw_expire));
#endif /* defined(HAVE_PASSWD_CHANGE) && defined(HAVE_PASSWD_EXPIRE) */

	/* Nothing else left to fail -- really log in. */

        /*
         * Update the utmp files, both BSD and SYSV style.
         */
        if (utmpx_login(tty, username, hostname ? hostname : "") != 0
	    && !fflag) {
                printf("No utmpx entry.  You must exec \"login\" from the lowest level \"sh\".\n");
                sleepexit(0);
        }
	utmp_login(ttyn, username, hostname ? hostname : "");
	dolastlog(quietlog);

	/*
	 * Set device protections, depending on what terminal the
	 * user is logged in. This feature is used on Suns to give
	 * console users better privacy.
	 */
	login_fbtab(tty, pwd->pw_uid, pwd->pw_gid);

	(void)chown(ttyn, pwd->pw_uid,
	    (gr = getgrnam(TTYGRPNAME)) ? gr->gr_gid : pwd->pw_gid);
	chmod(ttyn, S_IRUSR | S_IWUSR | S_IWGRP);
	(void)setgid(pwd->pw_gid);

	initgroups(username, pwd->pw_gid);

	if (*pwd->pw_shell == '\0')
		pwd->pw_shell = _PATH_BSHELL;

        /*
         * Set up a new environment. With SYSV, some variables are always
         * preserved; some varables are never preserved, and some variables
         * are always clobbered. With BSD, nothing is always preserved, and
         * some variables are always clobbered. We add code to make sure
         * that LD_* and IFS are never preserved.
         */
	if (term[0] == '\0')
		(void)strncpy(term, stypeof(tty), sizeof(term));
        /* set up a somewhat censored environment. */
        sysv_newenv(argc, argv, pwd, term, pflag);
#ifdef KERBEROS
	if (krbtkfile_env)
		(void)setenv("KRBTKFILE", krbtkfile_env, 1);
#endif

	if (tty[sizeof("tty")-1] == 'd')
		syslog(LOG_INFO, "DIALUP %s, %s", tty, pwd->pw_name);

	/* If fflag is on, assume caller/authenticator has logged root login. */
	if (rootlogin && fflag == 0)
		if (hostname)
			syslog(LOG_NOTICE, "ROOT LOGIN (%s) ON %s FROM %s",
			    username, tty, hostname);
		else
			syslog(LOG_NOTICE, "ROOT LOGIN (%s) ON %s", username, tty);

#ifdef KERBEROS
	if (!quietlog && notickets == 1 && !noticketsdontcomplain)
		(void)printf("Warning: no Kerberos tickets issued.\n");
#endif

#ifdef LOGALL
	/*
	 * Syslog each successful login, so we don't have to watch hundreds
	 * of wtmp or lastlogin files.
	 */
	if (hostname) {
		syslog(LOG_INFO, "login from %s as %s", hostname, pwd->pw_name);
	} else {
		syslog(LOG_INFO, "login on %s as %s", tty, pwd->pw_name);
	}
#endif

#ifndef NO_MOTD
        /*
         * Optionally show the message of the day. System V login leaves
         * motd and mail stuff up to the shell startup file.
         */
	if (!quietlog) {
	        struct stat st;
#if 0
		(void)printf("%s\n\t%s  %s\n\n",
	    "Copyright (c) 1980, 1983, 1986, 1988, 1990, 1991, 1993, 1994",
		    "The Regents of the University of California. ",
		    "All rights reserved.");
#endif
		motd();
		(void)sprintf(tbuf,
		    "%s/%s", _PATH_MAILDIR, pwd->pw_name);
		if (stat(tbuf, &st) == 0 && st.st_size != 0)
			(void)printf("You have %smail.\n",
			    (st.st_mtime > st.st_atime) ? "new " : "");
	}
#endif /* NO_MOTD */

#ifdef LOGIN_ACCESS
	if (login_access(pwd->pw_name, hostname ? full_hostname : tty) == 0) {
		printf("Permission denied\n");
		if (hostname)
			syslog(LOG_NOTICE, "%s LOGIN REFUSED FROM %s",
				pwd->pw_name, hostname);
		else
			syslog(LOG_NOTICE, "%s LOGIN REFUSED ON %s",
				pwd->pw_name, tty);
		sleepexit(1);
	}
#endif

	(void)signal(SIGALRM, SIG_DFL);
	(void)signal(SIGQUIT, SIG_DFL);
	(void)signal(SIGINT, SIG_DFL);
	(void)signal(SIGTSTP, SIG_IGN);

	tbuf[0] = '-';
	(void)strcpy(tbuf + 1, (p = strrchr(pwd->pw_shell, '/')) ?
	    p + 1 : pwd->pw_shell);

#ifdef HAVE_SETLOGIN
     	if (setlogin(pwd->pw_name) < 0)
                syslog(LOG_ERR, "setlogin() failure: %m");
#endif

#ifdef HAVE_SETPCRED
	if (setpcred (pwd->pw_name, NULL) == -1)
		syslog(LOG_ERR, "setpcred() failure: %m");
#endif /* HAVE_SETPCRED */

	/* Discard permissions last so can't get killed and drop core. */
	{
	    int uid = rootlogin ? 0 : pwd->pw_uid;
	    if(setuid(uid) != 0){
		fprintf(stderr, "Failed to setuid(%d): %s\n", 
			uid, strerror(errno));
		if(!rootlogin)
		    exit(1);
	    }
	}
		       

	/*
         * After dropping privileges and after cleaning up the environment,
         * optionally run, as the user, /bin/passwd.
         */
 
        if (pwd->pw_passwd[0] == 0 && strcasecmp(default_passreq, "YES") == 0) {
                printf("You don't have a password.  Choose one.\n");
                if (change_passwd(pwd))
                        sleepexit(0);
		changepass = 0;
        }

#ifdef SYSV_SHADOW
#ifdef HAVE_GETSPNAM
	spwd = getspnam (username);
#endif
        if (spwd && sysv_expire(spwd)) {
                if (change_passwd(pwd))
                        sleepexit(0);
		changepass = 0;
        }
#endif /* SYSV_SHADOW */
	if (changepass) {
		int res;
		if ((res=system(_PATH_CHPASS)))
			sleepexit(1);
	}

	if (k_hasafs()) {
		k_setpag();
		k_afsklog(0, 0);
	}

	execlp(pwd->pw_shell, tbuf, 0);
	if (getuid() == 0) {
		fprintf(stderr, "Can't exec %s, trying %s\n", 
			pwd->pw_shell, _PATH_BSHELL);
		execlp(_PATH_BSHELL, tbuf, 0);
		err(1, "%s", _PATH_BSHELL);
	}
	err(1, "%s", pwd->pw_shell);
}

#ifdef	KERBEROS
#define	NBUFSIZ		(UT_NAMESIZE + 1 + 5)	/* .root suffix */
#else
#define	NBUFSIZ		(UT_NAMESIZE + 1)
#endif

void
getloginname(prompt)
     int prompt;
{
	int ch;
	char *p;
	static char nbuf[NBUFSIZ];

	for (;;) {
                if (prompt)
                    if (ttyprompt && *ttyprompt)
                        (void)printf("%s", ttyprompt);
                    else
		        (void)printf("login: ");
		prompt = 1;
		for (p = nbuf; (ch = getchar()) != '\n'; ) {
			if (ch == EOF) {
				badlogin(username);
				exit(0);
			}
			if (p < nbuf + (NBUFSIZ - 1))
				*p++ = ch;
		}
		if (p > nbuf)
			if (nbuf[0] == '-')
				(void)fprintf(stderr,
				    "login names may not start with '-'.\n");
			else {
				*p = '\0';
				username = nbuf;
				break;
			}
	}
}

int
rootterm(ttyn)
	char *ttyn;
{
#ifndef HAVE_TTYENT_H
        return (default_console == 0 || strcmp(default_console, ttyname(0)) == 0);
#else
	struct ttyent *t;

	return ((t = getttynam(ttyn)) && t->ty_status & TTY_SECURE);
#endif
}

#ifndef NO_MOTD /* message of the day stuff */

jmp_buf motdinterrupt;

void
motd()
{
	int fd, nchars;
	RETSIGTYPE (*oldint)();
	char tbuf[8192];

	if ((fd = open(_PATH_MOTDFILE, O_RDONLY, 0)) < 0)
		return;
	oldint = signal(SIGINT, sigint);
	if (setjmp(motdinterrupt) == 0)
		while ((nchars = read(fd, tbuf, sizeof(tbuf))) > 0)
			(void)write(fileno(stdout), tbuf, nchars);
	(void)signal(SIGINT, oldint);
	(void)close(fd);
}

/* ARGSUSED */
void
sigint(signo)
	int signo;
{

	longjmp(motdinterrupt, 1);
}

#endif /* !NO_MOTD */

/* ARGSUSED */
void
timedout(signo)
	int signo;
{

	(void)fprintf(stderr, "Login timed out after %d seconds\n",
		      login_timeout);
	exit(0);
}

void
checknologin()
{
	int fd, nchars;
	char tbuf[8192];

	if ((fd = open(_PATH_NOLOGIN, O_RDONLY, 0)) >= 0) {
		while ((nchars = read(fd, tbuf, sizeof(tbuf))) > 0)
			(void)write(fileno(stdout), tbuf, nchars);
		sleepexit(0);
	}
}

void
dolastlog(quiet)
	int quiet;
{
#if defined(HAVE_LASTLOG_H) || defined(HAVE_LOGIN_H) || defined(SYSV_SHADOW)
	struct lastlog ll;
	int fd;

	if ((fd = open(_PATH_LASTLOG, O_RDWR, 0)) >= 0) {
		(void)lseek(fd, (off_t)pwd->pw_uid * sizeof(ll), SEEK_SET);
#ifdef SYSV_SHADOW
                if (read(fd, (char *)&ll, sizeof(ll)) == sizeof(ll) &&
                    ll.ll_time != 0) {
                        if (pwd->pw_uid && spwd && spwd->sp_inact > 0
                            && ll.ll_time / DAY + spwd->sp_inact < DAY_NOW) {
                                printf("Your account has been inactive too long.\n");
                                sleepexit(1);
                        }
                        if (!quiet) {
                                printf("Last login: %.*s ",
                                    24-5, (char *)ctime(&ll.ll_time));
                                if (*ll.ll_host != '\0') {
                                        printf("from %.*s\n",
                                            (int)sizeof(ll.ll_host),
					       ll.ll_host);
                                } else
                                        printf("on %.*s\n",
                                            (int)sizeof(ll.ll_line),
					       ll.ll_line);
                        }
                }
                (void)lseek(fd, (off_t)pwd->pw_uid * sizeof(ll), SEEK_SET);
#else /* SYSV_SHADOW */
		if (!quiet) {
			if (read(fd, (char *)&ll, sizeof(ll)) == sizeof(ll) &&
			    ll.ll_time != 0) {
				(void)printf("Last login: %.*s ",
				    24-5, (char *)ctime(&ll.ll_time));
				if (*ll.ll_host != '\0')
					(void)printf("from %.*s\n",
					    (int)sizeof(ll.ll_host),
					    ll.ll_host);
				else
					(void)printf("on %.*s\n",
					    (int)sizeof(ll.ll_line),
					    ll.ll_line);
			}
			(void)lseek(fd, (off_t)pwd->pw_uid * sizeof(ll), SEEK_SET);
		}
#endif /* SYSV_SHADOW */
		memset((void *)&ll, 0, sizeof(ll));
		(void)time(&ll.ll_time);
		(void)strncpy(ll.ll_line, tty, sizeof(ll.ll_line));
		if (hostname)
			(void)strncpy(ll.ll_host, hostname, sizeof(ll.ll_host));
		(void)write(fd, (char *)&ll, sizeof(ll));
		(void)close(fd);
	}
#endif /* DOLASTLOG */
}

void
badlogin(name)
	char *name;
{

	if (failures == 0)
		return;
	if (hostname) {
		syslog(LOG_NOTICE, "%d LOGIN FAILURE%s FROM %s",
		    failures, failures > 1 ? "S" : "", hostname);
		syslog(LOG_AUTHPRIV|LOG_NOTICE,
		    "%d LOGIN FAILURE%s FROM %s, %s",
		    failures, failures > 1 ? "S" : "", hostname, name);
	} else {
		syslog(LOG_NOTICE, "%d LOGIN FAILURE%s ON %s",
		    failures, failures > 1 ? "S" : "", tty);
		syslog(LOG_AUTHPRIV|LOG_NOTICE,
		    "%d LOGIN FAILURE%s ON %s, %s",
		    failures, failures > 1 ? "S" : "", tty, name);
	}
}

#undef	UNKNOWN
#define	UNKNOWN	"su"

char *
stypeof(ttyid)
        char *ttyid;
{
    /* TERM is probably a better guess than anything else. */
    char *term = getenv("TERM");

    if (term != 0 && term[0] != 0)
	return term;

    {
#ifndef HAVE_TTYENT_H
	return UNKNOWN;
#else
	struct ttyent *t;
	return (ttyid && (t = getttynam(ttyid)) ? t->ty_type : UNKNOWN);
#endif
    }
}

static
void
xgetstr(char *buf, int cnt, char *err)
{
        char ch;
 
        do {
                if (read(0, &ch, sizeof(ch)) != sizeof(ch))
                        exit(1);
                if (--cnt < 0) {
                        fprintf(stderr, "%s too long\r\n", err);
                        sleepexit(1);
                }
                *buf++ = ch;
        } while (ch);
}

/*
 * Some old rlogind's unknowingly pass remuser, locuser and
 * terminal_type/speed so we need to take care of that part of the
 * protocol here. Also, we can't make a getpeername(2) on the socket
 * so we have to trust that rlogind resolved the name correctly.
 */

static
int
doremotelogin(char *host)
{
        int code;
        char *cp;

        xgetstr(rusername, sizeof (rusername), "remuser");
        xgetstr(lusername, sizeof (lusername), "locuser");
        xgetstr(term, sizeof(term), "Terminal type");
	cp = strchr(term, '/');
	if (cp != 0)
		*cp = 0;	/* For now ignore speed/bg */
        pwd = k_getpwnam(lusername);
        if (pwd == NULL)
                return(-1);
        code = ruserok(host, (pwd->pw_uid == 0), rusername, lusername);
	if (code == 0)
	  syslog(LOG_NOTICE,
		 "Warning: An old rlogind accepted login probably from host %s",
		 host);
	return(code);
}
 
void
sleepexit(int eval)
{

	(void)sleep(5);
	exit(eval);
}

